/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// Include files

#include <iomanip>
#include <iostream>
#include <set>

// local
#include "DiagSolvToolBase.h"

//-----------------------------------------------------------------------------
// Implementation file for class : DiagSolvTool
//
// 2007-06 : Adlene Hicheur
//-----------------------------------------------------------------------------

// Declaration of the Tool Factory
// DECLARE_COMPONENT( DiagSolvTool )

//=============================================================================
// Initialize
//=============================================================================
StatusCode DiagSolvToolBase::initialize() {
  StatusCode sc = GaudiTool::initialize();
  info() << "EigenValueThreshold = " << m_eigenValueThreshold << endmsg;
  info() << "MinEigenModeChisquare = " << m_minEigenModeChisquare << endmsg;
  return sc;
}

namespace {
  struct SortByAbsSecond {
    template <class T>
    bool operator()( const T& lhs, const T& rhs ) const {
      return std::abs( lhs.second ) < std::abs( rhs.second );
    }
  };
} // namespace

StatusCode DiagSolvToolBase::compute( AlSymMat& bigmatrix, AlVec& bigvector,
                                      IAlignSolvTool::SolutionInfo& solinfo ) const {
  info() << "Solving with diagonalization method" << endmsg;

  // declare transition matrix + vector to store eigenvalues
  const size_t N = bigmatrix.rows();
  AlMat        U( N, N );
  AlMat        V( N, N );
  AlVec        S( N );

  debug() << "Before diagonalization" << endmsg;

  // defer the diagonalization to whatever tool derives from this.
  // bigmatrix = U S V^T
  StatusCode sc = diagonalize( bigmatrix, U, V, S );
  if ( !sc.isSuccess() ) {
    error() << "Diagonalization failed" << endmsg;
    return sc;
  }

  debug() << "After diagonalization" << endmsg;

  //   Compute bigvector in diagonal basis. Here we need the transpose of U, just like it would be in GSL.
  const Eigen::VectorXd D = U.transpose() * bigvector;

  // Perform some analysis of the eigenvalues.
  solinfo.numNegativeEigenvalues = solinfo.numSmallEigenvalues = 0;
  {
    // count the number of negative eigenvalues and number smaller than one
    std::vector<std::pair<size_t, double>> sortedev( N );
    for ( size_t ipar = 0; ipar < N; ++ipar ) {
      sortedev[ipar] = std::make_pair( ipar, S[ipar] );
      if ( std::abs( S[ipar] ) < 1.0 ) ++solinfo.numSmallEigenvalues;

      // Eigen and Gsl put the 'sign' of negative eigenvalues in a
      // difference between U and V. This explains why we do not see
      // negative Eigenvalues and need to use both U and V in the
      // solution.
      const double sign = U.col( ipar ).dot( V.col( ipar ) );
      if ( !( sign * S[ipar] > 0.0 ) ) {
        ++solinfo.numNegativeEigenvalues;
        // Issue a warning for each negative eigenvalue.
        warning() << "Negative eigenvalue (i,val,chi2)= " << ipar << " " << sign * S( ipar ) << " "
                  << D( ipar ) * D( ipar ) / S( ipar ) << endmsg;
      }
    }

    // Dump the 10 smallest eigenvalues. We'll sort these really in
    // abs, ignoring the sign.
    std::sort( sortedev.begin(), sortedev.end(), SortByAbsSecond() );
    std::ostringstream logmessage;
    logmessage << "Smallest eigen values: [ " << std::setprecision( 4 );
    size_t maxpar = std::min( N, m_numberOfPrintedEigenvalues.value() );
    for ( size_t ipar = 0; ipar < maxpar; ++ipar )
      logmessage << sortedev[ipar].second << ( ( ipar < maxpar - 1 ) ? ", " : " ]" );

    // consider the smallest eigenvalue
    if ( N > 0 ) {
      // dump the corresponding eigenvector
      if ( N <= m_numberOfPrintedEigenvalues ) {
        logmessage << std::endl;
        logmessage << "Eigenvector for smallest eigenvalue: [ ";
        for ( size_t ipar = 0; ipar < N; ++ipar ) {
          logmessage << U( sortedev[0].first, ipar ) << ( ( ipar < N - 1 ) ? ", " : " ]" );
        }
      }
      // figure out which parameter contributes most to this eigenvector
      size_t thepar( 0 );
      for ( size_t ipar = 0; ipar < N; ++ipar ) {
        if ( std::abs( U( sortedev[0].first, thepar ) ) < std::abs( U( sortedev[0].first, ipar ) ) ) thepar = ipar;
      }
      solinfo.minEigenValue         = sortedev[0].second;
      solinfo.weakestModeMaxPar     = thepar;
      solinfo.weakestModeMaxParCoef = U( sortedev[0].first, thepar );
    }
    info() << logmessage.str() << endmsg;
  }

  // compute alignment corrections and their variance. first flag
  // modes that we cut away by eigenvalue. (constraints have large
  // negative value, so the 'abs' should do)
  std::vector<bool> keepEigenValue( N, true );
  if ( m_eigenValueThreshold > 0 )
    for ( size_t i = 0; i < N; i++ )
      keepEigenValue[i] = std::abs( S( i ) ) > m_eigenValueThreshold ||
                          ( m_minEigenModeChisquare > 0 && D( i ) * D( i ) / S( i ) > m_minEigenModeChisquare );

  double                sumchisqaccepted( 0 ), sumchisqrejected( 0 ), maxchisq( 0 );
  size_t                numrejected( 0 );
  std::multiset<double> chisqvalues;
  for ( size_t i = 0; i < N; i++ ) {
    double thischisq = D( i ) * D( i ) / S( i );
    if ( !keepEigenValue[i] ) {
      info() << "Rejecting eigenvalue: val = " << S( i ) << " chisq = " << thischisq << endmsg;
      ++numrejected;
      sumchisqrejected += thischisq;
    } else {
      chisqvalues.insert( thischisq );
      if ( maxchisq < thischisq ) maxchisq = thischisq;
      sumchisqaccepted += thischisq;
      if ( std::abs( S( i ) ) < m_eigenValueThreshold )
        info() << "Accepting low eigenvalue: val = " << S( i ) << " chisq = " << thischisq << endmsg;
    }
  }

  solinfo.totalChisq = sumchisqaccepted;

  info() << "Number / total chi2 of rejected eigenvalues: " << numrejected << " " << sumchisqrejected << endmsg;
  info() << "Total chi2 of accepted eigenvalues: " << sumchisqaccepted << endmsg;
  info() << "Maximum chi2 of individual mode: " << maxchisq << endmsg;

  // because the lagrange constraints lead to negative ev, they also
  // give negative chi2 contributions. that's all fine for the total
  // chi2, but it leads to problems if you want to identify the
  // 'largest' chi2 contribution: such contributions may be
  // compensated by large negative contributions. Let's solve this
  // pragmatically:
  // - make a multiset of all chisq contributions
  // - while the smallest one is negative, remove the smallest and
  //   largest from the set, and reinsert the sum
  // - the largest remaining eigenvalue is the one we need;
  while ( !chisqvalues.empty() && *( chisqvalues.begin() ) < 0 && *( chisqvalues.rbegin() ) > 0 ) {
    double a = *( chisqvalues.begin() );
    chisqvalues.erase( chisqvalues.begin() );
    double b = *( chisqvalues.rbegin() );
    chisqvalues.erase( --( chisqvalues.rbegin().base() ) );
    chisqvalues.insert( a + b );
  }
  if ( !chisqvalues.empty() ) solinfo.maxChisqEigenMode = *( chisqvalues.rbegin() );

  info() << "Maximum chi2 of individual mode after correcting for lagrange constraints: " << solinfo.maxChisqEigenMode
         << endmsg;

  // reset the input
  for ( size_t i = 0; i < N; i++ ) {
    bigvector( i ) = 0.;
    for ( size_t j = 0; j <= i; j++ ) bigmatrix( i, j ) = 0.;
  }

  // now fill it
  for ( size_t k = 0; k < N; ++k )
    if ( keepEigenValue[k] ) {
      for ( size_t i = 0; i < N; ++i ) {
        bigvector( i ) += V( i, k ) * D( k ) / S( k );
        for ( size_t j = 0; j <= i; ++j ) bigmatrix( i, j ) += V( i, k ) * U( j, k ) / S( k );
      }
    } else {
      warning() << "Rejecting eigenvalue: " << k << " " << S( k ) << endmsg;
    }

  return sc;
}
