/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
//-----------------------------------------------------------------------------
// Implementation file for class : AlignSaveTuple
//
// Louis Nicolas, Adlene Hicheur (EPFL)
// Started: 06/06
///-----------------------------------------------------------------------------

//===========================================================================
// Includes
//===========================================================================
// Local
#include "AlignSaveTuple.h"

#include "GaudiKernel/IMagneticFieldSvc.h"

// Interfaces
#include "MCInterfaces/ITrackGhostClassification.h"
#include "TrackInterfaces/ITrackExtrapolator.h"
#include "TrackKernel/TrackFunctors.h"

// Event
#include "Event/GhostTrackInfo.h"

//===========================================================================

//===========================================================================
// Declare
//===========================================================================
// Declaration of the algorithm factory
DECLARE_COMPONENT( AlignSaveTuple )
//===========================================================================

//===========================================================================
// Constructor
//===========================================================================
AlignSaveTuple::AlignSaveTuple( const std::string& name, ISvcLocator* pSvcLocator )
    : GaudiTupleAlg( name, pSvcLocator ), m_extrapolator( "TrackParabolicExtrapolator" ) {

  // Reserving space for some vectors
  m_sharedHits.reserve( 1000 );

  // Ghost classification
  this->declareProperty( "GhostClassification", m_ghostToolName = "TTrackGhostClassification" );

  this->declareProperty( "TracksLocation", m_tracksPath = "Rec/Track/Best" );

  // Are we using MC or real data?
  this->declareProperty( "MCData", m_mcData = true );

  // 48 = 12 (layers) * 2 (ladder overlaps) * 2 (box overlaps)
  this->declareProperty( "IsolatedTrackNStripsTolerance", m_nStripsTol = 2 );
  this->declareProperty( "IsolatedTrackNStrawsTolerance", m_nStrawsTol = 1 );

  this->declareProperty( "MaxNHits", m_maxNHits = 40 );
}
//===========================================================================

//===========================================================================
// Destructor
//===========================================================================
AlignSaveTuple::~AlignSaveTuple() {}
//===========================================================================

//===========================================================================
// Initialize
//===========================================================================
StatusCode AlignSaveTuple::initialize() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Initialize" << endmsg;

  StatusCode sc = GaudiTupleAlg::initialize(); // must be executed first
  if ( sc.isFailure() ) return Error( "Failed to initialize" );

  // The extrapolator
  m_extrapolator.retrieve().ignore();

  // Retrieve the ghost classification tool
  m_ghostClassification = tool<ITrackGhostClassification>( m_ghostToolName, "GhostTool", this );

  // Get The Magnetic Field
  m_pIMF   = svc<IMagneticFieldSvc>( "MagneticFieldSvc", true );
  m_iEvent = 0;

  debug() << "AlignSaveTuple initialized successfully" << endmsg;

  return StatusCode::SUCCESS;
}
//===========================================================================

//===========================================================================
// Execute
//===========================================================================
StatusCode AlignSaveTuple::execute() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "AlignSaveTuple starting execution" << endmsg;

  // Get the odin information
  odin = get<LHCb::ODIN>( LHCb::ODINLocation::Default );

  // Get the tracks
  m_tracks = get<LHCb::Tracks>( m_tracksPath );

  // Get the hits shared by more than one track
  m_sharedHits.clear();
  getSharedHits();

  // Creating the tuple in memory
  Tuples::Tuple trackSelTuple = nTuple( "AllVariables" );

  // Get the association table
  AsctTool associator( evtSvc(), m_tracksPath );
  m_directTable = associator.direct();

  //**********************************************************************
  // Global Variables
  //**********************************************************************
  m_runNr = odin->runNumber();
  m_evtNr = odin->eventNumber();
  ++m_iEvent;
  m_eventMultiplicity = m_tracks->size();
  //**********************************************************************

  if ( msgLevel( MSG::DEBUG ) ) { debug() << "Retrieved " << m_eventMultiplicity << " tracks." << endmsg; }

  // Loop over tracks - select some and make some plots
  LHCb::Tracks::const_iterator iTracks = m_tracks->begin();
  m_iGoodTrack                         = 0;
  for ( ; iTracks != m_tracks->end(); ++iTracks ) {
    LHCb::Track& aTrack = **iTracks;

    if ( m_mcData ) {
      if ( iTracks == m_tracks->begin() ) {
        if ( isGhostTrack( &aTrack ) )
          m_ghostRate = (double)1 / m_eventMultiplicity;
        else
          m_ghostRate = 0;
      }

      LHCb::Tracks::const_iterator iTracks2 = iTracks + 1;
      for ( ; iTracks2 != m_tracks->end(); ++iTracks2 ) {
        LHCb::Track& tr2 = **iTracks2;
        // Compute ghost rate
        if ( iTracks == m_tracks->begin() )
          if ( isGhostTrack( &tr2 ) ) m_ghostRate += (double)1 / m_eventMultiplicity;
      }
    }

    if ( msgLevel( MSG::DEBUG ) )
      debug() << "******************************************************" << endmsg << "Entering new good track"
              << endmsg;

    // Get the number of IT and OT hits
    StatusCode sc;
    if ( ( sc = fillVariables( &aTrack, trackSelTuple ) ) == StatusCode::FAILURE ) return sc;
    ++m_iGoodTrack;
  }

  return StatusCode::SUCCESS;
}
//===========================================================================

//===========================================================================
// Finalize
//===========================================================================
StatusCode AlignSaveTuple::finalize() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Finalize" << endmsg;

  return GaudiTupleAlg::finalize(); // Must be called after all other actions
}
//===========================================================================

//===========================================================================
// Fill Variables
//===========================================================================
StatusCode AlignSaveTuple::fillVariables( const LHCb::Track* aTrack, Tuples::Tuple trackSelTuple ) {

  //**********************************************************************
  // Tracks Variables
  //**********************************************************************
  // Ghost? ==> ghost classification
  if ( m_mcData ) {
    m_isGhostTrack             = isGhostTrack( aTrack );
    m_ghostTrackClassification = defValue;
    LHCb::GhostTrackInfo ghostClassInfo;
    if ( m_isGhostTrack ) {
      m_ghostClassification->info( *aTrack, ghostClassInfo ).ignore();
      m_ghostTrackClassification = ghostClassInfo.classification();
      if ( msgLevel( MSG::DEBUG ) ) debug() << "Ghost classification = " << ghostClassInfo.classification() << endmsg;
    }
  }
  // Ghost Track Classification:
  // 0  = None
  // 1  = Real
  // 10 = Conversion
  // 11 = EM
  // 12 = FromPhi
  // 13 = HadronicInteraction
  // 14 = DecayInFlight
  // 20 = SpilloverAndNoise
  // 21 = GhostParent
  // 22 = InconsistentParts
  // 23 = Combinatoric

  m_nSharedHits = 0;
  m_fSharedHits = 0;
  m_nHoles      = 0;

  m_trackChi2PerDoF   = aTrack->chi2PerDoF();
  m_trackChi2Prob     = aTrack->probChi2();
  m_trackFitMatchChi2 = aTrack->info( LHCb::Track::AdditionalInfo::FitMatchChi2, 999999 );

  m_trackEta = aTrack->pseudoRapidity();

  m_trackP    = aTrack->p() / Gaudi::Units::GeV;
  m_trackPt   = aTrack->pt() / Gaudi::Units::GeV;
  m_trackErrP = sqrt( aTrack->firstState().errP2() ) / Gaudi::Units::GeV;
  if ( m_mcData ) {
    m_trackMCP  = defValue;
    m_trackMCPt = defValue;
  }

  // Track Slope and Position at 7500 mm
  LHCb::State aState = aTrack->closestState( 7500. );
  m_entryTX          = aState.tx();
  m_entryTY          = aState.ty();
  m_entryX           = aState.x();
  m_entryY           = aState.y();
  //**********************************************************************

  //**********************************************************************
  // Hits Variables
  //**********************************************************************
  m_res.clear();
  m_errRes.clear();

  m_hitX.clear();
  m_hitY.clear();
  m_hitZ.clear();

  m_nLadOverlaps = 0;
  m_ladOvlapRes.clear();
  m_ladOvlapLay.clear();

  m_nBoxOverlaps = 0;
  m_boxOvlapRes.clear();
  m_boxOvlapSta.clear();
  //**********************************************************************

  if ( m_mcData ) {
    //**********************************************************************
    // MCP and MCPt (get MCParticle linked to the track)
    //**********************************************************************
    // Retrieve the Linker table
    LinkedTo<LHCb::MCParticle, LHCb::Track> directLink( evtSvc(), msgSvc(), m_tracksPath );
    if ( directLink.notFound() ) Error( "Linker table not found", StatusCode::SUCCESS, 1 ).ignore();

    // Get MCParticle linked by highest weight to Track and get its p and pt
    LHCb::MCParticle* mcPart = new LHCb::MCParticle;
    mcPart                   = directLink.first( aTrack );
    if ( 0 != mcPart ) {
      m_trackMCP  = mcPart->p() / Gaudi::Units::GeV;
      m_trackMCPt = mcPart->pt() / Gaudi::Units::GeV;
    }
    delete mcPart;
    //**********************************************************************
  }

  std::vector<LHCb::LHCbID>::const_iterator iIDs = aTrack->lhcbIDs().begin();
  for ( ; iIDs != aTrack->lhcbIDs().end(); ++iIDs ) {
    const LHCb::LHCbID& aID = *iIDs;
    //**********************************************************************
    // Number of shared hits in the track
    //**********************************************************************
    if ( std::binary_search( m_sharedHits.begin(), m_sharedHits.end(), aID, lessByID() ) ) ++m_nSharedHits;
    //**********************************************************************
  }

  // Loop over the nodes to get the hits variables
  const auto& nodes_ = nodes( *aTrack );
  for ( auto iNodes = nodes_.begin(); iNodes != nodes_.end(); ++iNodes ) {

    const LHCb::FitNode* fNode = *iNodes;

    // Only loop on hits with measurement
    if ( !fNode->hasMeasurement() ) continue;

    //**********************************************************************
    // Residuals and Residual Errors
    //**********************************************************************
    //     m_res.push_back(aNode.residual());
    //     m_errRes.push_back(aNode.errResidual());
    m_res.push_back( fNode->unbiasedResidual() );
    m_errRes.push_back( fNode->errUnbiasedResidual() );
    //**********************************************************************

    //**********************************************************************
    // Hit Position
    //**********************************************************************
    Gaudi::XYZPoint hitPos = fNode->state().position();
    m_hitX.push_back( hitPos.X() );
    m_hitY.push_back( hitPos.Y() );
    m_hitZ.push_back( hitPos.Z() );
    //**********************************************************************
  }

  //**********************************************************************
  // Printing some informations in debug mode:
  //**********************************************************************
  if ( msgLevel( MSG::DEBUG ) )
    debug() << "This track has " << 0 << " IT expected hits" << endmsg << "               " << m_nLadOverlaps
            << " IT ladder overlaps" << endmsg << "               " << m_nBoxOverlaps << " IT box overlaps" << endmsg
            << "               " << m_nHoles << " holes" << endmsg << "               " << m_nSharedHits
            << " shared hits"
            << " (= " << 100 * m_fSharedHits << " %)" << endmsg;
  //**********************************************************************

  return writeNTuple( trackSelTuple );
}
//===========================================================================

//===========================================================================
// Is track ghost or not?
//===========================================================================
bool AlignSaveTuple::isGhostTrack( const LHCb::Track* aTrack ) {

  DirectRange range;
  if ( !m_directTable ) {
    Error( "Track MCParticle table missing. Are you running TrackAssociator before?" ).ignore();
    return false;
  }
  range = m_directTable->relations( aTrack );

  return ( range.empty() ? true : false );
}
//===========================================================================

//===========================================================================
// Check if this hit is shared with (at least) one other track
//===========================================================================
void AlignSaveTuple::getSharedHits() {

  LHCb::Tracks::const_iterator iTracks = m_tracks->begin();
  for ( ; iTracks != m_tracks->end(); ++iTracks ) {
    const LHCb::Track*                        aTrack = *iTracks;
    std::vector<LHCb::LHCbID>::const_iterator iIDs   = aTrack->lhcbIDs().begin();
    for ( ; iIDs != aTrack->lhcbIDs().end(); ++iIDs ) {
      const LHCb::LHCbID aID = *iIDs;
      if ( isSharedHit( aTrack, aID ) ) m_sharedHits.push_back( aID );
    }
  }

  // sorting and stripping out duplicates
  std::sort( m_sharedHits.begin(), m_sharedHits.end(), lessByID() );
  m_sharedHits.erase( std::unique( m_sharedHits.begin(), m_sharedHits.end() ), m_sharedHits.end() );
}
//===========================================================================

//===========================================================================
// Check if this hit is shared with (at least) one other track
//===========================================================================
bool AlignSaveTuple::isSharedHit( const LHCb::Track* theTrack, const LHCb::LHCbID aID ) {

  LHCb::Tracks::const_iterator iTracks = m_tracks->begin();
  for ( ; iTracks != m_tracks->end(); ++iTracks ) {
    const LHCb::Track* aTrack = *iTracks;
    if ( aTrack != theTrack )
      if ( aTrack->isOnTrack( aID ) ) {
        if ( msgLevel( MSG::DEBUG ) ) debug() << "This hit is being shared" << endmsg;
        return true;
      }
  }

  return false;
}
//===========================================================================

//===========================================================================
// Write NTuple
//===========================================================================
StatusCode AlignSaveTuple::writeNTuple( Tuples::Tuple trackSelTuple ) {

  //**********************************************************************
  // Fill the NTuple columns
  //**********************************************************************
  //     plot(variable, tag, "variable", min, max, nbins);
  // Global Variables
  trackSelTuple->column( "IEvent", m_iEvent ).ignore();
  trackSelTuple->column( "EvtNr", m_evtNr ).ignore();
  trackSelTuple->column( "RunNr", m_runNr ).ignore();
  trackSelTuple->column( "Multiplicity", m_eventMultiplicity ).ignore();
  if ( m_mcData ) trackSelTuple->column( "GhostRate", m_ghostRate ).ignore();

  trackSelTuple->column( "NVeLoClusters", 0 ).ignore();
  trackSelTuple->column( "NITClusters", 0 ).ignore();

  // Track Variables
  trackSelTuple->column( "ITrack", m_iGoodTrack ).ignore();

  if ( m_mcData ) {
    trackSelTuple->column( "IsAGhost", m_isGhostTrack ).ignore();
    trackSelTuple->column( "GhostTrackClassification", m_ghostTrackClassification ).ignore();
  }

  trackSelTuple->column( "NITHits", 0 ).ignore();
  trackSelTuple->column( "NOTHits", 0 ).ignore();
  trackSelTuple->column( "NSharedHits", m_nSharedHits ).ignore();
  trackSelTuple->column( "FSharedHits", m_fSharedHits ).ignore();
  trackSelTuple->column( "NNeighbouringHits", 0 ).ignore();
  trackSelTuple->column( "NHoles", m_nHoles ).ignore();

  trackSelTuple->column( "TrackChi2PerDoF", m_trackChi2PerDoF ).ignore();
  trackSelTuple->column( "TrackChi2Prob", m_trackChi2Prob ).ignore();
  trackSelTuple->column( "TrackFitMatchChi2", m_trackFitMatchChi2 ).ignore();

  trackSelTuple->column( "TrackEta", m_trackEta ).ignore();

  trackSelTuple->column( "TrackP", m_trackP ).ignore();
  trackSelTuple->column( "TrackPt", m_trackPt ).ignore();
  trackSelTuple->column( "TrackErrP", m_trackErrP ).ignore();
  if ( m_mcData ) {
    trackSelTuple->column( "TrackMCP", m_trackMCP ).ignore();
    trackSelTuple->column( "TrackMCPt", m_trackMCPt ).ignore();
  }
  // Track State closest to 7500. mm
  trackSelTuple->column( "TXAt7500", m_entryTX ).ignore();
  trackSelTuple->column( "TYAt7500", m_entryTY ).ignore();
  trackSelTuple->column( "XAt7500", m_entryX ).ignore();
  trackSelTuple->column( "YAt7500", m_entryY ).ignore();

  // Overlaps
  trackSelTuple
      ->farray( "LadOvlapRes", m_ladOvlapRes.begin(), m_ladOvlapRes.end(), "NLadOverlaps", int( m_maxNHits / 2 ) )
      .ignore();
  trackSelTuple
      ->farray( "LadOvlapLay", m_ladOvlapLay.begin(), m_ladOvlapLay.end(), "NLadOverlaps", int( m_maxNHits / 2 ) )
      .ignore();

  trackSelTuple->farray( "BoxOvlapRes", m_boxOvlapRes.begin(), m_boxOvlapRes.end(), "NBoxOverlaps", m_nStations )
      .ignore();
  trackSelTuple->farray( "BoxOvlapStation", m_boxOvlapSta.begin(), m_boxOvlapSta.end(), "NBoxOverlaps", m_nStations )
      .ignore();

  // Hits Variables
  trackSelTuple->farray( "Residuals", m_res.begin(), m_res.end(), "NResiduals", m_maxNHits ).ignore();
  trackSelTuple->farray( "ErrResiduals", m_errRes.begin(), m_errRes.end(), "NResiduals", m_maxNHits ).ignore();

  trackSelTuple->farray( "HitX", m_hitX.begin(), m_hitX.end(), "NResiduals", m_maxNHits ).ignore();
  trackSelTuple->farray( "HitY", m_hitY.begin(), m_hitY.end(), "NResiduals", m_maxNHits ).ignore();
  trackSelTuple->farray( "HitZ", m_hitZ.begin(), m_hitZ.end(), "NResiduals", m_maxNHits ).ignore();
  //**********************************************************************

  return trackSelTuple->write();
}
//===========================================================================
