/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#ifndef _AlignSaveTuple_H_
#define _AlignSaveTuple_H_

/** @class AlignSaveTuple AlignSaveTuple Align/AlignTrTools/AlignSaveTuple.h
 *
 *  Provide a monitoring tool for T-Station alignment studies
 *
 *  @author L. Nicolas
 *  @date   13/06/2006
 */

//===========================================================================
// Includes
//===========================================================================
// Gaudi
#include "GaudiAlg/GaudiTupleAlg.h"
#include "GaudiKernel/ToolHandle.h"

// Linkers
#include "Linker/LinkedTo.h"
#include "Linker/LinkerTool.h"

// Tools
#include "TrackInterfaces/IHitExpectation.h"
#include "TrackInterfaces/ITrackExtrapolator.h"

// Event
#include "Event/FitNode.h"
#include "Event/MCParticle.h"
#include "Event/ODIN.h"
#include "Event/Track.h"
//===========================================================================

//===========================================================================
// Forward declarations
//===========================================================================
struct ITrackExtrapolator;
class IMagneticFieldSvc;
struct ITrackGhostClassification;
class GhostTrackInfo;
//===========================================================================

class AlignSaveTuple : public GaudiTupleAlg {
public:
  // Constructors and destructor
  AlignSaveTuple( const std::string& name, ISvcLocator* pSvcLocator );
  virtual ~AlignSaveTuple();

  StatusCode initialize() override;
  StatusCode execute() override;
  StatusCode finalize() override;

private:
  typedef std::vector<double>                       Array;
  typedef LinkerTool<LHCb::Track, LHCb::MCParticle> AsctTool;
  typedef AsctTool::DirectType                      DirectTable;
  typedef DirectTable::Range                        DirectRange;

  static const int defValue = -999999;

  // Fill the Variables
  StatusCode fillVariables( const LHCb::Track* aTrack, Tuples::Tuple trackSelTuple );

  // Is the track linked to a MC particle or not?
  bool isGhostTrack( const LHCb::Track* aTrack );

  // Check if the hit is being shared with (at least) one other Track
  void getSharedHits();
  bool isSharedHit( const LHCb::Track* aTrack, const LHCb::LHCbID aID );

  // Write the NTuple
  StatusCode writeNTuple( Tuples::Tuple trackSelTuple );

  //======================================================================
  // Properties
  //======================================================================
  int m_nStations;

  int m_maxNHits;

  double m_nStripsTol;
  double m_nStrawsTol;

  std::string m_tracksPath;

  std::string m_ghostToolName;
  //======================================================================

  ToolHandle<ITrackExtrapolator> m_extrapolator;        // Interface to track extrapolator
  IMagneticFieldSvc*             m_pIMF;                ///< Pointer to the magn. field service
  ITrackGhostClassification*     m_ghostClassification; ///< Pointer to ghost tool

  const LHCb::Tracks*               m_tracks;
  mutable std::vector<LHCb::LHCbID> m_sharedHits;
  const LHCb::ODIN*                 odin;

  DirectTable* m_directTable;
  bool         m_mcData;

  // Global Variables
  long   m_iEvent;
  long   m_evtNr;
  long   m_runNr;
  int    m_eventMultiplicity;
  double m_ghostRate;

  // Track Variables
  int m_iGoodTrack;

  bool m_isGhostTrack;
  int  m_ghostTrackClassification;

  int    m_nSharedHits;
  double m_fSharedHits;
  int    m_nHoles;

  double m_trackChi2PerDoF;
  double m_trackChi2Prob;
  double m_trackFitMatchChi2;

  double m_trackEta;

  double m_trackP;
  double m_trackPt;
  double m_trackErrP;
  double m_trackMCP;
  double m_trackMCPt;

  double m_entryTX;
  double m_entryTY;
  double m_entryX;
  double m_entryY;

  // Hits Variables
  Array m_res;
  Array m_errRes;

  Array m_hitX;
  Array m_hitY;
  Array m_hitZ;

  int   m_nLadOverlaps;
  Array m_ladOvlapRes;
  Array m_ladOvlapLay;

  int   m_nBoxOverlaps;
  Array m_boxOvlapRes;
  Array m_boxOvlapSta;

  //=============================================================================
  // for comparing LHCbIDs
  //=============================================================================
  class lessByID {
  public:
    inline bool operator()( const LHCb::LHCbID& obj1, const LHCb::LHCbID& obj2 ) const {
      return obj1.lhcbID() < obj2.lhcbID();
    }
  };
  //=============================================================================
};

#endif // _AlignSaveTuple_H_
