/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// $Id: AlignTrackMonitor.cpp,v 1.17 2009-10-09 12:00:00 wouter Exp $
//

//-----------------------------------------------------------------------------
// Implementation file for class : AlignTrackMonitor
//
// Louis Nicolas, Adlene Hicheur (EPFL)
// Started: 06/06
///-----------------------------------------------------------------------------

//===========================================================================
// Includes
//===========================================================================
// Local
#include "AlignTrackMonitor.h"

#include "GaudiKernel/IMagneticFieldSvc.h"
#include "GaudiKernel/ToStream.h"

// Interfaces
#include "TrackInterfaces/ITrackExtrapolator.h"
#include "TrackKernel/TrackFunctors.h"

//===========================================================================

//===========================================================================
// Declare
//===========================================================================
// Declaration of the algorithm factory
DECLARE_COMPONENT( AlignTrackMonitor )
//===========================================================================

//===========================================================================
// Constructor
//===========================================================================
AlignTrackMonitor::AlignTrackMonitor( const std::string& name, ISvcLocator* pSvcLocator )
    : GaudiHistoAlg( name, pSvcLocator ), m_extrapolator( "TrackParabolicExtrapolator" ) {

  // Reserving space for some vectors
  m_sharedHits.reserve( 1000 );

  // Location of the different objects in use
  this->declareProperty( "TracksLocation", m_tracksPath = "Rec/Track/Best" );

  // 48 = 12 (layers) * 2 (ladder overlaps) * 2 (box overlaps)
  this->declareProperty( "IsolatedTrackNStripsTolerance", m_nStripsTol = 2 );
  this->declareProperty( "IsolatedTrackNStrawsTolerance", m_nStrawsTol = 1 );
}
//===========================================================================

//===========================================================================
// Initialize
//===========================================================================
StatusCode AlignTrackMonitor::initialize() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Initialize" << endmsg;

  StatusCode sc = GaudiHistoAlg::initialize(); // must be executed first
  if ( sc.isFailure() ) return Error( "Failed to initialize" );

  // Get tools
  m_extrapolator.retrieve().ignore();

  // Get The Magnetic Field
  m_pIMF = svc<IMagneticFieldSvc>( "MagneticFieldSvc", true );

  debug() << "AlignTrackMonitor initialized successfully" << endmsg;

  return StatusCode::SUCCESS;
}
//===========================================================================

//===========================================================================
// Execute
//===========================================================================
StatusCode AlignTrackMonitor::execute() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "AlignTrackMonitor starting execution" << endmsg;

  // Get the tracks
  m_tracks = get<LHCb::Tracks>( m_tracksPath );

  // Get the hits shared by more than one track
  m_sharedHits.clear();
  getSharedHits();

  //**********************************************************************
  // Global Variables
  //**********************************************************************
  m_eventMultiplicity = m_tracks->size();
  //**********************************************************************

  if ( msgLevel( MSG::DEBUG ) ) { debug() << "Retrieved " << m_eventMultiplicity << " tracks." << endmsg; }

  // Loop over tracks - select some and make some plots
  LHCb::Tracks::const_iterator iTracks = m_tracks->begin();
  for ( ; iTracks != m_tracks->end(); ++iTracks ) {
    LHCb::Track& aTrack = **iTracks;

    if ( msgLevel( MSG::DEBUG ) )
      debug() << "******************************************************" << endmsg << "Entering new good track"
              << endmsg;

    StatusCode sc;
    if ( ( sc = fillVariables( &aTrack ) ) == StatusCode::FAILURE ) return sc;
  }

  return StatusCode::SUCCESS;
}
//===========================================================================

//===========================================================================
// Finalize
//===========================================================================
StatusCode AlignTrackMonitor::finalize() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Finalize" << endmsg;

  return GaudiHistoAlg::finalize(); // Must be called after all other actions
}
//===========================================================================

//===========================================================================
// Fill Variables
//===========================================================================
StatusCode AlignTrackMonitor::fillVariables( const LHCb::Track* aTrack ) {

  //**********************************************************************
  // Tracks Variables
  //**********************************************************************
  m_nSharedHits  = 0;
  m_fSharedHits  = 0;
  m_nLadOverlaps = 0;
  m_nBoxOverlaps = 0;

  m_trackChi2PerDoF = aTrack->chi2PerDoF();
  m_trackChi2Prob   = aTrack->probChi2();

  // Get momentum, transverse momentum and error in GeV
  m_trackP    = aTrack->p() / Gaudi::Units::GeV;
  m_trackPt   = aTrack->pt() / Gaudi::Units::GeV;
  m_trackErrP = sqrt( aTrack->firstState().errP2() ) / Gaudi::Units::GeV;

  // Track pseudo-rapidity
  m_trackEta = aTrack->pseudoRapidity();

  // Track Slope and Position at 7500 mm
  LHCb::State aState = aTrack->closestState( 7500. );
  m_entryTX          = aState.tx();
  m_entryTY          = aState.ty();
  m_entryX           = aState.x();
  m_entryY           = aState.y();
  //**********************************************************************

  // Loop over the nodes to get the hits variables
  const auto& nodes_ = nodes( *aTrack );
  for ( auto iNodes = nodes_.begin(); iNodes != nodes_.end(); ++iNodes ) {

    // Only loop on hits with measurement
    if ( !( **iNodes ).hasMeasurement() ) continue;

    const LHCb::FitNode& aNode  = **iNodes;
    LHCb::LHCbID         nodeID = aNode.measurement().lhcbID();

    //**********************************************************************
    // Number of shared hits in the track
    //**********************************************************************
    if ( std::binary_search( m_sharedHits.begin(), m_sharedHits.end(), nodeID, lessByID() ) ) ++m_nSharedHits;
    //**********************************************************************
  }

  //**********************************************************************
  // Printing some informations in debug mode:
  //**********************************************************************
  if ( msgLevel( MSG::DEBUG ) )
    debug() << "This track has " << 0 << " IT hits" << endmsg << "               " << 0 << " IT expected hits" << endmsg
            << "               " << m_nLadOverlaps << " IT ladder overlaps" << endmsg << "               " << 0
            << " IT box overlaps" << endmsg << "               " << 0 << " OT hits" << endmsg << "               " << 0
            << " OT expected hits" << endmsg << "               " << 0 << " holes" << endmsg << "               "
            << m_nSharedHits << " shared hits"
            << " (= " << 100 * m_fSharedHits << " %)" << endmsg << "               " << 0 << " neighbouring hits"
            << endmsg;
  //**********************************************************************

  return fillHistos();
}
//===========================================================================

//===========================================================================
// Check if this hit is shared with (at least) one other track
//===========================================================================
void AlignTrackMonitor::getSharedHits() {

  LHCb::Tracks::const_iterator iTracks = m_tracks->begin();
  for ( ; iTracks != m_tracks->end(); ++iTracks ) {
    const LHCb::Track*                        aTrack = *iTracks;
    std::vector<LHCb::LHCbID>::const_iterator iIDs   = aTrack->lhcbIDs().begin();
    for ( ; iIDs != aTrack->lhcbIDs().end(); ++iIDs ) {
      const LHCb::LHCbID aID = *iIDs;
      if ( isSharedHit( aTrack, aID ) ) m_sharedHits.push_back( aID );
    }
  }

  // sorting and stripping out duplicates
  std::sort( m_sharedHits.begin(), m_sharedHits.end(), lessByID() );
  m_sharedHits.erase( std::unique( m_sharedHits.begin(), m_sharedHits.end() ), m_sharedHits.end() );
}
//===========================================================================

//===========================================================================
// Check if this hit is shared with (at least) one other track
//===========================================================================
bool AlignTrackMonitor::isSharedHit( const LHCb::Track* theTrack, const LHCb::LHCbID aID ) {

  LHCb::Tracks::const_iterator iTracks = m_tracks->begin();
  for ( ; iTracks != m_tracks->end(); ++iTracks ) {
    const LHCb::Track* aTrack = *iTracks;
    if ( aTrack != theTrack )
      if ( aTrack->isOnTrack( aID ) ) {
        if ( msgLevel( MSG::DEBUG ) ) debug() << "This hit is being shared" << endmsg;
        return true;
      }
  }

  return false;
}
//===========================================================================

//===========================================================================
// Fill the histos
//===========================================================================
StatusCode AlignTrackMonitor::fillHistos() {

  // Event Variables
  plot( m_eventMultiplicity, "Multiplicity", "Multiplicity", 0., 500., 50 );

  // Various numbers related to the track
  plot( m_nSharedHits, "NSharedHits", "NSharedHits", -0.5, 30.5, 31 );
  plot( m_fSharedHits, "FSharedHits", "FSharedHits", -0.01, 1.01, 50 );
  plot( m_nLadOverlaps, "NLadOverlaps", "NLadOverlaps", -0.5, 15.5, 16 );
  plot( m_nBoxOverlaps, "NBoxOverlaps", "NBoxOverlaps", -0.5, 3.5, 4 );

  // Track fit "quality"
  plot( m_trackChi2PerDoF, "TrackChi2PerDoF", "TrackChi2PerDoF", -2., 100., 50 );
  plot( m_trackChi2Prob, "TrackChi2Prob", "TrackChi2Prob", -0.01, 1.01, 50 );

  // Track momentum and more
  plot( m_trackP, "TrackP", "TrackP", -5., 205., 50 );
  plot( m_trackPt, "TrackPt", "TrackPt", -0.1, 10.1, 50 );
  plot( m_trackErrP, "TrackErrP", "TrackErrP", -0.1, 10.1, 50 );

  // Track pseudo-rapidity
  plot( m_trackEta, "TrackEta", "TrackEta", 0., 10., 50 );

  // Track State closest to 7500. mm
  plot( m_entryTX, "TXAt7500", "TXAt7500", -1., 1., 50 );
  plot( m_entryTY, "TYAt7500", "TYAt7500", -0.5, 0.5, 50 );
  plot( m_entryX, "XAt7500", "XAt7500", -3000., 3000., 50 );
  plot( m_entryY, "YAt7500", "YAt7500", -3000., 3000., 50 );
  //**********************************************************************

  return StatusCode::SUCCESS;
}
//===========================================================================
