/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#ifndef _AlignTrackMonitor_H_
#define _AlignTrackMonitor_H_

/** @class AlignTrackMonitor AlignTrackMonitor Align/AlignTrTools/AlignTrackMonitor.h
 *
 *  Provide a monitoring tool for T-Station alignment studies
 *
 *  @author L. Nicolas
 *  @date   13/06/2006
 */

//===========================================================================
// Includes
//===========================================================================
// Gaudi
#include "GaudiAlg/GaudiHistoAlg.h"
#include "GaudiKernel/ToolHandle.h"

// Linkers
#include "Linker/LinkedTo.h"
#include "Linker/LinkerTool.h"

// Event
#include "Event/FitNode.h"
#include "Event/Track.h"

// Interface
#include "TrackInterfaces/IHitExpectation.h"
#include "TrackInterfaces/ITrackExtrapolator.h"

//===========================================================================

//===========================================================================
// Forward declarations
//===========================================================================
struct ITrackExtrapolator;
class IMagneticFieldSvc;
//===========================================================================

class AlignTrackMonitor : public GaudiHistoAlg {
public:
  // Constructors and destructor
  AlignTrackMonitor( const std::string& name, ISvcLocator* pSvcLocator );

  StatusCode initialize() override;
  StatusCode execute() override;
  StatusCode finalize() override;

private:
  typedef std::vector<double> Array;

  static const int defValue = -999999;

  // Fill the Variables
  StatusCode fillVariables( const LHCb::Track* aTrack );

  // Check if the hit is being shared with (at least) one other Track
  void getSharedHits();
  bool isSharedHit( const LHCb::Track* aTrack, const LHCb::LHCbID aID );

  // Fill the histos
  StatusCode fillHistos();

  //======================================================================
  // Properties
  //======================================================================
  std::string m_tracksPath;

  double m_nStripsTol;
  double m_nStrawsTol;
  //======================================================================

  ToolHandle<ITrackExtrapolator> m_extrapolator; // Interface to track extrapolator

  IMagneticFieldSvc* m_pIMF; ///< Pointer to the magn. field service

  // Various containers
  const LHCb::Tracks*               m_tracks;
  mutable std::vector<LHCb::LHCbID> m_sharedHits;

  // event variables
  int m_eventMultiplicity;

  // Various counters
  unsigned int m_nSharedHits;
  double       m_fSharedHits;
  unsigned int m_nLadOverlaps;
  unsigned int m_nBoxOverlaps;

  // Track "quality"
  double m_trackChi2PerDoF;
  double m_trackChi2Prob;

  // Track momentum and more
  double m_trackP;
  double m_trackPt;
  double m_trackErrP;

  // Track pseudo-rapidity
  double m_trackEta;

  // Track entry point in T-station
  double m_entryTX;
  double m_entryTY;
  double m_entryX;
  double m_entryY;

  //=============================================================================
  // for comparing LHCbIDs
  //=============================================================================
  class lessByID {
  public:
    inline bool operator()( const LHCb::LHCbID& obj1, const LHCb::LHCbID& obj2 ) const {
      return obj1.lhcbID() < obj2.lhcbID();
    }
  };
  //=============================================================================
};

#endif // _AlignTrackMonitor_H_
