/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

// Include files
// from std
#include <string>
#include <utility>
#include <vector>

// from Gaudi
#include "Gaudi/Accumulators.h"
#include "GaudiAlg/GaudiHistoAlg.h"
#include "GaudiKernel/GenericMatrixTypes.h"
#include "GaudiKernel/IUpdateable.h"
#include "GaudiKernel/Point3DTypes.h"
#include "GaudiKernel/ToolHandle.h"

// from TrackEvent
#include "Event/FitNode.h"
#include "Event/Particle.h"
#include "Event/RecVertex.h"
#include "Event/Track.h"

// from TrackInterfaces
#include "TrackInterfaces/ITrackProjector.h"
#include "TrackInterfaces/ITrackProjectorSelector.h"
#include "TrackInterfaces/ITrackSelector.h"

// from AlignSolvTools
#include "AlignKernel/AlSymMat.h"
#include "AlignKernel/AlVec.h"

// from AlignmentInterfaces
#include "AlignmentInterfaces/IAlignSolvTool.h"
#include "AlignmentInterfaces/IAlignUpdateTool.h"
#include "AlignmentInterfaces/IWriteAlignmentConditionsTool.h"
//#include "AlignmentInterfaces/IWriteMultiAlignmentConditionsTool.h"

// AIDA
#include "AIDA/IHistogram1D.h"
#include "AIDA/IHistogram2D.h"
#include "AIDA/IProfile1D.h"

// from local
// @todo: this should be moved to the interfaces package
#include "AlElementHistos.h"
#include "AlignmentElement.h"
#include "IAlignConstraintTool.h"
#include "IGetElementsToBeAligned.h"
#include "ITrackResidualTool.h"
#include "IVertexResidualTool.h"

namespace Al {
  class Equations;
  class Residuals;
} // namespace Al

/** @class AlignAlgorithm AlignAlgorithm.h
 *
 *
 *  @author Jan Amoraal
 *  @date   2007-03-05
 */

class HistoUpdater {
public:
  unsigned long m_Reference;
  IMonitorSvc*  m_MonSvc;
  void          setMonitorService( IMonitorSvc* ms ) { m_MonSvc = ms; }
  HistoUpdater() : m_Reference( 0 ), m_MonSvc( 0 ) {}
  unsigned long readReference( const std::string& refFileName ) {
    FILE* f = fopen( refFileName.c_str(), "r" );
    fscanf( f, "%lu", &m_Reference );
    fclose( f );
    printf( "[INFO]==================> Read Reference %ld from file %s\n", m_Reference, refFileName.c_str() );
    fflush( ::stdout );
    return m_Reference;
  }
  unsigned long getReference() { return m_Reference; }
  void          Update( int runnr, const std::string& refFileName ) {
    unsigned long ref = readReference( refFileName );
    if ( m_MonSvc != 0 ) {
      SmartIF<IUpdateableIF> aaa( m_MonSvc );
      if ( aaa.isValid() ) aaa->update( runnr * 100 + ref ).ignore();
    }
  }
};

class AlignAlgorithm : public GaudiHistoAlg {
public:
  /// Some handy typedefs
  typedef IGetElementsToBeAligned::Elements Elements;
  typedef std::vector<double>               AlignConstants;
  typedef std::vector<LHCb::FitNode*>       Nodes;
  typedef Gaudi::Matrix1x6                  Derivatives;
  typedef Gaudi::Matrix6x6                  HMatrix;
  enum { CanonicalConstraintOff = 0, CanonicalConstraintOn, CanonicalConstraintAuto } CanonicalConstraintStrategy;

  /// Standard constructor
  using GaudiHistoAlg::GaudiHistoAlg;

  StatusCode initialize() override; ///< Algorithm initialization
  StatusCode finalize() override;   ///< Algorithm initialization
  StatusCode execute() override;    ///< Algorithm execution
  StatusCode stop() override;       ///< Transition executed upon the Stop command
  StatusCode start() override;      ///< Transition executed upon the Start command

  /// Methods to call when an update is triggered
  void                 update();
  void                 update( const Al::Equations& equations );
  const Al::Equations* equations() const { return m_equations; }

  void reset();

protected:
  bool printDebug() const { return msgLevel( MSG::DEBUG ); };
  bool printVerbose() const { return msgLevel( MSG::VERBOSE ); };
  bool accumulate( const Al::Residuals& residuals );
  void resetHistos();

  typedef std::vector<LHCb::RecVertex>    VertexContainer;
  typedef std::vector<const LHCb::Track*> TrackContainer;
  void selectVertexTracks( const LHCb::RecVertex& vertex, const TrackContainer& tracks,
                           TrackContainer& tracksinvertex ) const;
  void removeVertexTracks( const LHCb::RecVertex& vertex, TrackContainer& tracks ) const;
  void removeParticleTracks( const LHCb::Particle& p, TrackContainer& tracks ) const;
  void splitVertex( const LHCb::RecVertex& vertex, const TrackContainer& tracks, VertexContainer& splitvertices ) const;
  LHCb::RecVertex* cloneVertex( const LHCb::RecVertex& vertex, const TrackContainer& selectedtracks ) const;
  bool             testNodes( const LHCb::Track& track ) const;

private:
  // properties
  Gaudi::Property<size_t> m_nIterations{this, "NumberOfIterations", 10u, "Number of iterations used on histogram axes"};
  PublicToolHandle<IGetElementsToBeAligned> m_elementtool{
      this, "AlignmentElements", "GetElementsToBeAligned"}; ///< Pointer to tool to align detector
  Gaudi::Property<std::string> m_tracksLocation{this, "TracksLocation", LHCb::TrackLocation::Default};
  Gaudi::Property<std::string> m_vertexLocation{this, "VertexLocation"}; ///< Vertices for alignment
  Gaudi::Property<std::string> m_particleLocation{this,
                                                  "ParticleLocation"}; ///< particles with mass constraint for alignment
  Gaudi::Property<std::string> m_histoPartitionName{this, "PartitionName", "LHCbA"};
  Gaudi::Property<std::string> m_histoRefFileName{this, "ReferenceFile", ""};

  ToolHandle<ITrackProjectorSelector>       m_projSelector{this, "ProjectorSelector",
                                                     "TrackProjectorSelector"}; ///< Pointer to projector selector tool
  PublicToolHandle<Al::ITrackResidualTool>  m_trackresidualtool{this, "TrackResidualTool", "Al::TrackResidualTool"};
  PublicToolHandle<Al::IVertexResidualTool> m_vertexresidualtool{this, "VertexResidualTool", "Al::VertexResidualTool"};
  PublicToolHandle<Al::IAlignUpdateTool>    m_updatetool{this, "UpdateTool", "Al::AlignUpdateTool"};
  PublicToolHandle<IWriteAlignmentConditionsTool> m_xmlWritersTool{this, "XmlWritersTool",
                                                                   "WriteMultiAlignmentConditionsTool"};
  Gaudi::Property<std::string>                    m_alignSummaryDataSvc{this, "AlignSummaryDataSvc", "DetectorDataSvc"};
  Gaudi::Property<std::string> m_alignSummaryLocation{this, "AlignSummaryLocation", "AlignDerivativeData"};
  Gaudi::Property<bool>        m_fillHistos{this, "FillHistos", false};

  Gaudi::Property<bool>        m_correlation{this, "UseCorrelations",
                                      true}; ///< Do we take into account correlations between residuals?
  Gaudi::Property<bool>        m_updateInFinalize{this, "UpdateInFinalize", false}; ///< Call update from finalize
  Gaudi::Property<double>      m_chi2Outlier{this, "Chi2Outlier", 1e4}; ///< Chi2 threshold for outlier rejection
  Gaudi::Property<size_t>      m_minTracksPerVertex{this, "MinTracksPerVertex", 2};
  Gaudi::Property<size_t>      m_maxTracksPerVertex{this, "MaxTracksPerVertex", 8};
  Gaudi::Property<std::string> m_outputDataFileName{this, "OutputDataFile"};
  Gaudi::Property<std::vector<std::string>> m_inputDataFileNames{this, "InputDataFiles"};
  Gaudi::Property<long long>                m_forcedInitTime{
      this, "ForcedInitialTime", 0,
      "force the alignment geometry to initialize with this time (rather than first event)"};
  Gaudi::Property<bool>                     m_Online{this, "OnlineMode", false};
  Gaudi::Property<std::vector<std::string>> m_RunList{this, "RunList"};

private:
  mutable Gaudi::Accumulators::Counter<> m_iteration{this, "Iteration"};
  mutable Gaudi::Accumulators::Counter<> m_nTracks{this, "Number of tracks used"};
  mutable Gaudi::Accumulators::Counter<> m_covFailure{this, "Number of covariance calculation failures"};

  Al::Equations* m_equations{nullptr}; ///< Equations to solve
  bool           m_resetHistos{false}; // reset histos on next event processing
  HistoUpdater   m_HistoUpdater;

  /// Monitoring
  // @todo: Move this to a monitoring tool
  std::vector<AlElementHistos> m_elemHistos;
};
