/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// Include files

// from STL
#include <algorithm>
#include <iomanip>

// from Gaudi
#include "GaudiKernel/DataObject.h"
#include "GaudiKernel/IDataProviderSvc.h"
#include "GaudiKernel/IDetDataSvc.h"
#include "GaudiKernel/IUpdateManagerSvc.h"

// from TrackEvent
#include "Event/FitNode.h"
#include "Event/Measurement.h"

// from Det
#include "FTDet/DeFTDetector.h"
#include "MuonDet/DeMuonDetector.h"
#include "UTDet/DeUTDetector.h"
#include "VPDet/DeVP.h"

// from Boost
#include "boost/regex.hpp"

// // local
#include "AlignKernel/AlEquations.h"
#include "GetElementsToBeAligned.h"

//-----------------------------------------------------------------------------
// Implementation file for class : GetElementsToBeAligned
//
// 2007-10-08 : Jan Amoraal
//-----------------------------------------------------------------------------

// Declaration of the Tool Factory
DECLARE_COMPONENT( GetElementsToBeAligned )

StatusCode GetElementsToBeAligned::initialize() {

  StatusCode sc = GaudiTool::initialize();
  if ( sc.isFailure() ) return Error( "Failed to initialize tool!", sc );

  if ( m_elemsToBeAligned.empty() ) { warning() << "You haven't specified any elements to align!" << endmsg; }

  info() << "Use local frame = " << m_useLocalFrame << endmsg;

  // delay initialization till the first event
  m_initTime = 0;

  /// Get pointer to detector svc
  IDataProviderSvc* detectorSvc = 0;
  sc                            = service( "DetectorDataSvc", detectorSvc, true );
  if ( sc.isFailure() ) {
    error() << "Something went wrong with retrieving detector data svc" << endmsg;
    return sc;
  }

  /// Retrieve LHCb detector in Transient store
  const std::string lhcbDetectorPath = "/dd/Structure/LHCb";
  DataObject*       lhcbObj          = 0;
  sc                                 = detectorSvc->retrieveObject( lhcbDetectorPath, lhcbObj );
  if ( sc.isFailure() ) {
    error() << "==> Couldn't retrieve LHCb Detector from transient store!" << endmsg;
    detectorSvc->release();
    return sc;
  }

  const IDetectorElement* lhcb = dynamic_cast<const IDetectorElement*>( lhcbObj );
  /// This shouldn't go wrong, but you never know
  if ( !lhcb ) {
    error() << "==> Object is not a detector element!" << endmsg;
    detectorSvc->release();
    return StatusCode::FAILURE;
  }

  info() << "===================== GetElementsToAlign =====================" << endmsg;
  debug() << "   Using the following regular expressions: " << endmsg;

  size_t                                 index( 0 );
  typedef std::vector<AlignmentElement*> NonConstElements;
  NonConstElements                       alignelements;
  for ( std::vector<std::string>::const_iterator i = m_elemsToBeAligned.begin(), iEnd = m_elemsToBeAligned.end();
        i != iEnd; ++i ) {
    /// Split string into path to det elem regex and dofs to align
    bool        groupElems = false;
    std::string groupname, path, dofs;

    // first : second : third
    // tokens of strings
    std::vector<std::string> tokens;
    Tokenizer                split( ( *i ), Separator( " :" ) );
    for ( Tokenizer::iterator j = split.begin(), jEnd = split.end(); j != jEnd; ++j ) tokens.push_back( ( *j ) );

    if ( tokens.size() == 2 ) {
      path = tokens.at( 0 );
      dofs = tokens.at( 1 );
    } else if ( tokens.size() == 3 ) {
      groupElems = true;
      path       = tokens.at( 1 );
      dofs       = tokens.at( 2 );
      groupname  = tokens.at( 0 );
      if ( groupname.find( "Group" ) != std::string::npos ) groupname = path;
    } else {
      error() << "==> There is something wrong with the specified property Elements: " << tokens.size() << std::endl
              << *i << endmsg;
      return StatusCode::FAILURE;
    }

    if ( msgLevel( MSG::DEBUG ) ) {
      debug() << "Group elements   : " << groupElems << endmsg;
      if ( groupElems ) debug() << "Group name       : " << groupname << endmsg;
      debug() << "Path to detelems : " << path << endmsg;
      debug() << "DoFs to align    : " << dofs << endmsg;
    }

    // Traverse LHCb detector in transient store and get alignable elements
    std::vector<const DetectorElement*> detelements;
    // Break path into a set of regexs. Forward slash is the path separator
    Tokenizer elemtokenizer( path, Separator( "/" ) );
    RegExs    regexs;

    // Print the list of regular expressions that are going to be used.
    debug() << "       ";
    for ( Tokenizer::iterator j = elemtokenizer.begin(), jEnd = elemtokenizer.end(); j != jEnd; ++j )
      debug() << "\"" << ( *j ) << "\""
              << " ";
    debug() << endmsg;

    // Create list of regular expressions
    for ( Tokenizer::iterator j = elemtokenizer.begin(), jEnd = elemtokenizer.end(); j != jEnd; ++j ) {
      boost::regex ex;
      // Check if token is a valid regular expression, else catch exception and return statuscode failure.
      // Better to stop the program and let the user fix the expression than trying to predict what he/she wants.
      try {
        ex.assign( ( *j ), boost::regex_constants::icase );
      } catch ( boost::bad_expression& exprs ) {
        error() << "==> Error: " << ( *j ) << " is not a valid regular expression: " << exprs.what() << endmsg;
        return StatusCode::FAILURE;
      }
      // If valid add expression to list of expressions.
      regexs.push_back( ex );
    }

    // Depth is equal to the number regular expressions in the regex list.
    getElements( lhcb, regexs, regexs.size(), detelements );

    // Check that we have found elements to align, else exit gracefully
    if ( detelements.empty() ) {
      error() << "\n ==> Couldn't find any elements that matched the given regular expression! \n"
              << " expression=\'" << *i << "\'" << std::endl
              << " ==> The syntax of the property Elements is a list of : \"Group : Regex representing path of det "
                 "elems : dofs\" \n"
              << " ==> where group and dofs are optional.\n"
              << " ==> Check the regular expression and also check if there are no whitespaces.\n"
              << endmsg;
      return StatusCode::FAILURE;
    }

    // Loop over elements and create AlignmentElements
    if ( groupElems ) {
      // first check that there isn't already a group with this name. if there is, add the elements.
      NonConstElements::iterator ielem = alignelements.begin();
      while ( ielem != alignelements.end() && ( *ielem )->name() != groupname ) ++ielem;
      if ( ielem != alignelements.end() ) {
        ( *ielem )->addElements( detelements );
        ( *ielem )->addDofs( dofs );
      } else
        alignelements.push_back( new AlignmentElement( groupname, detelements, index++, dofs, m_useLocalFrame ) );
    } else {
      for ( std::vector<const DetectorElement*>::iterator ielem = detelements.begin(); ielem != detelements.end();
            ++ielem ) {
        // check that there isn't already a group with this name. if there is, only set the dofs
        std::string                name  = AlignmentElement::stripElementName( ( *ielem )->name() );
        NonConstElements::iterator jelem = alignelements.begin();
        while ( jelem != alignelements.end() && ( *jelem )->name() != name ) ++jelem;
        if ( jelem != alignelements.end() ) {
          warning() << "Multiple specifications of dofs for alignable " << name << ". Using '" << dofs << "'."
                    << endmsg;
          ( *jelem )->setDofs( dofs );
        } else {
          alignelements.push_back( new AlignmentElement( **ielem, index++, dofs, m_useLocalFrame ) );
        }
      }
    }
  }

  // sort the elements by hierarchy. currently, this just follows the
  // name of the first associated detector element.
  std::stable_sort( alignelements.begin(), alignelements.end(), ElementSorter() );

  // make sure to reset the indices after sorting
  index = 0;
  for ( NonConstElements::iterator ielem = alignelements.begin(); ielem != alignelements.end(); ++ielem, ++index )
    ( *ielem )->setIndex( index );

  // set up the mother-daughter relations. This only works once the
  // elements are sorted. We start with the lowest level elements, and
  // work our way up the tree.
  for ( NonConstElements::reverse_iterator ielem = alignelements.rbegin(); alignelements.rend() != ielem; ++ielem ) {
    bool                               found = false;
    NonConstElements::reverse_iterator jelem = ielem;
    for ( ++jelem; jelem != alignelements.rend() && !found; ++jelem )
      // is 'i' a daughter of 'j'
      if ( ( *jelem )->isOffspring( **ielem ) ) {
        ( const_cast<AlignmentElement*>( ( *jelem ) ) )->addDaughter( **ielem );
        found = true;
      }
  }

  // now fill the element map. this map is used to assign a certain
  // hit to a certain alignment element. again, it is important that
  // alignment elements are sorted: we add the hit only to the lowest
  // level element in the tree.
  for ( NonConstElements::const_iterator ielem = alignelements.begin(); ielem != alignelements.end(); ++ielem ) {
    AlignmentElement::ElementContainer allelements = ( *ielem )->elementsInTree();
    for ( AlignmentElement::ElementContainer::const_iterator idetelem = allelements.begin();
          idetelem != allelements.end(); ++idetelem )
      m_elementMap[*idetelem] = *ielem;
  }

  // copy the non-const elements into the local container.
  m_elements.insert( m_elements.end(), alignelements.begin(), alignelements.end() );

  /// Print list of detector elements that are going to be aligned

  info() << "   Going to align " << m_elements.size() << " detector elements:" << endmsg;
  info() << "   Number of elements in map: " << m_elementMap.size() << endmsg;

  info() << "==============================================================" << endmsg;

  return sc;
}

StatusCode GetElementsToBeAligned::finalize() {
  for ( Elements::const_iterator ialignable = m_elements.begin(); ialignable != m_elements.end(); ++ialignable )
    delete *ialignable;
  m_elements.clear();
  return GaudiTool::finalize();
}

void GetElementsToBeAligned::getElements( const IDetectorElement* parent, const RegExs& regexs, size_t depth,
                                          std::vector<const DetectorElement*>& detelements ) const {
  /// loop over children
  for ( IDetIter iC = parent->childBegin(), iCend = parent->childEnd(); iC != iCend; ++iC ) {
    /// Get path of child
    const std::string childPath = ( *iC )->name();
    if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Element  = " << childPath << endmsg;

    /// forward slash is path seperator
    Separator sep( "/" );
    /// Tokenise path
    Tokenizer tokenizer( childPath, sep );

    std::list<std::string> paths;
    /// Copy tokens to a temporary list of sub paths. This allows us to determine how deep we are.
    std::copy( tokenizer.begin(), tokenizer.end(), std::back_inserter( paths ) );

    const size_t currentDepth = paths.size();

    bool match = true; ///< Can we match a sub path to a regular expression?

    /// Begin iterator of regular expressions list
    std::list<boost::regex>::const_iterator iR = regexs.begin();
    /// End iterator of regular expressions list
    std::list<boost::regex>::const_iterator iRend = regexs.end();
    /// Loop over list of sub paths and try to match them
    for ( std::list<std::string>::const_iterator i = paths.begin(), iEnd = paths.end(); i != iEnd && iR != iRend;
          ++i ) {
      match = boost::regex_match( ( *i ), ( *iR++ ) );
      if ( !match ) break; /// If it doesn't match break loop. No need to try and match the rest
    }

    /// OK we found a detector element
    if ( match ) {
      if ( currentDepth == depth ) { detelements.push_back( dynamic_cast<DetectorElement*>( *iC ) ); }
      /// Call thyself. Get children recursively.
      /// No need to go any deeper.
      if ( paths.size() < depth ) getElements( ( *iC ), regexs, depth, detelements );
    }
  }
}

StatusCode GetElementsToBeAligned::findElements( const std::string&                    path,
                                                 std::vector<const AlignmentElement*>& alignelements ) const {
  alignelements.clear();
  std::copy_if( m_elements.begin(), m_elements.end(), std::back_inserter( alignelements ),
                [&, ex = boost::regex{path, boost::regex_constants::icase}]( const auto* ialelem ) {
                  return ialelem->name() == path || boost::regex_match( ialelem->name(), ex ) ||
                         // if any detector element matches, this is also fine
                         std::any_of( ialelem->detelements().begin(), ialelem->detelements().end(),
                                      [&]( const auto* elem ) { return boost::regex_match( elem->name(), ex ); } );
                } );
  return StatusCode::SUCCESS;
}

const AlignmentElement* GetElementsToBeAligned::findElement( const LHCb::FitNode& node ) const {
  const AlignmentElement* elem = nullptr;
  if ( node.hasMeasurement() ) {
    const LHCb::Measurement& meas    = node.measurement();
    const DetectorElement*   detelem = meas.detectorElement();
    if ( detelem ) {
      elem = findElement( *detelem );
    } else {
      elem = findElement( meas.lhcbID() );
    }
  }
  return elem;
}

const AlignmentElement* GetElementsToBeAligned::findElement( const LHCb::Measurement& meas ) const {
  const DetectorElement* detelem = meas.detectorElement();
  return detelem ? findElement( *detelem ) : findElement( meas.lhcbID() );
}

const AlignmentElement* GetElementsToBeAligned::findElement( const DetectorElement& element ) const {
  auto it = m_elementMap.find( &element );
  return it == m_elementMap.end() ? nullptr : it->second;
}

const AlignmentElement* GetElementsToBeAligned::findElement( const LHCb::LHCbID& id ) const {
  // first get the detector element from the measurement. this is no
  // longer needed if we have a new version of brunel.
  const DetectorElement* element( nullptr ); // = meas.detectorElement() ;
  {
    static DeVP*         vp = nullptr;
    static DeUTDetector* ut = nullptr;
    static DeFTDetector* ft = nullptr;

    vp                          = getDet<DeVP>( DeVPLocation::Default );
    ut                          = getDet<DeUTDetector>( DeUTDetLocation::UT );
    ft                          = getDet<DeFTDetector>( DeFTDetectorLocation::Default );
    static DeMuonDetector* muon = getDet<DeMuonDetector>( DeMuonLocation::Default );

    switch ( id.detectorType() ) {
    case LHCb::LHCbID::VP: {
      element = &vp->sensor( id.vpID().sensor() );
    } break;
    case LHCb::LHCbID::UT:
      element = ut->findSector( id.utID() );
      break;
    case LHCb::LHCbID::FT:
      element = ft->findModule( id.ftID() );
      break;
    case LHCb::LHCbID::Muon:
      element = muon->getChmbPtr( id.muonID().station(), id.muonID().region(),
                                  ( muon->Tile2Chamber( id.muonID() ) )[0]->chamberNumber() );
      break;
    default:
      error() << "LHCbID from non-tracking detector" << endmsg;
    }
  }

  // see if that element is in the map
  auto it = m_elementMap.find( element );
  return it == m_elementMap.end() ? nullptr : it->second;
}

void GetElementsToBeAligned::initEquations( Al::Equations& equations ) const {
  equations = Al::Equations( m_elements.size(), initTime() );
  for ( const auto& ielem : m_elements ) {
    // set the current delta (called alpha in Al::Equations)
    auto& elemdata = equations.element( ielem->index() );
    elemdata.setAlpha( ielem->currentDelta().transformParameters() );
    elemdata.setAlignFrame( AlParameters( ielem->alignmentFrame() ).transformParameters() );
  }
}

StatusCode GetElementsToBeAligned::initAlignmentFrame( Gaudi::Time now ) {
  StatusCode sc = StatusCode::SUCCESS;
  {
    std::stringstream message, message_dbg;
    message << std::left << std::setw( 80u ) << std::setfill( '*' ) << std::endl;
    IDetDataSvc* detDataSvc( 0 );
    sc = service( "DetectorDataSvc", detDataSvc, false );
    if ( sc.isFailure() ) {
      error() << "Could not retrieve DetectorDataSvc" << endmsg;
      return sc;
    }
    if ( now == 0 ) {
      now = detDataSvc->eventTime();
    } else {
      message << "Updating detector data svc and update manager svc with time: " << now << std::endl;
      IUpdateManagerSvc* updateMgrSvc( 0 );
      sc = service( "UpdateManagerSvc", updateMgrSvc, false );
      if ( sc.isFailure() ) {
        error() << "Could not retrieve UpdateManagerSvc" << endmsg;
        return sc;
      }
      detDataSvc->setEventTime( now );
      updateMgrSvc->newEvent().ignore();
    }

    m_initTime = now;
    message << "Initializing alignment frames with time: " << m_initTime.ns() << " --> "
            << m_initTime.format( true, "%F %r" );
    // update all elements with this time
    for ( Elements::iterator ielem = m_elements.begin(); ielem != m_elements.end(); ++ielem ) {
      try {
        ( const_cast<AlignmentElement*>( *ielem ) )->initAlignmentFrame();
      } catch ( std::string& errorMsg ) { error() << errorMsg; }
    }

    // now print the positions, just for debugging
    for ( Elements::const_iterator i = m_elements.begin(); i != m_elements.end(); ++i ) message_dbg << **i;
    message_dbg << std::left << std::setw( 80u ) << std::setfill( '*' ) << "";

    info() << message.str() << endmsg;
    debug() << message_dbg.str() << endmsg;
  }
  return sc;
}
