/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

// Include files
// from STL
#include <list>
#include <map>
#include <string>
#include <utility>
#include <vector>

// from Gaudi
#include "GaudiAlg/GaudiTool.h"

// Interface
#include "IGetElementsToBeAligned.h"

// from BOOST
#include "boost/regex.hpp"
#include "boost/tokenizer.hpp"

/** @class GetElementsToBeAligned GetElementsToBeAligned.h
 *
 *
 *  @author Jan Amoraal
 *  @date   2007-10-08
 */

class GetElementsToBeAligned : public extends<GaudiTool, IGetElementsToBeAligned> {
public:
  typedef IDetectorElement::IDEContainer::const_iterator IDetIter;  ///< IDetector element iterator
  typedef boost::char_separator<char>                    Separator; ///< Char separator
  typedef boost::tokenizer<Separator>                    Tokenizer; ///< List of tokens
  typedef std::list<boost::regex>                        RegExs;    ///< List of regular expressions

  /// Standard constructor
  using extends::extends;

  StatusCode initialize() override;
  StatusCode finalize() override;

  // Virtual in IGetElementsToBeAligned
  // Return pair of begin iter and end iter
  const Elements& elements() const override { return m_elements; }

  // Virtual in IGetElementsToBeAligned
  // Return method that finds an alignment element for a given LHCb id
  const AlignmentElement* findElement( const LHCb::LHCbID& anLHCbID ) const override;

  // Return method that finds an alignment element for a given Measuerment
  const AlignmentElement* findElement( const LHCb::FitNode& node ) const override;

  // Return method that finds an alignment element for a given Measuerment
  const AlignmentElement* findElement( const LHCb::Measurement& meas ) const override;

  // Return method that finds an alignment element for a given detector element
  const AlignmentElement* findElement( const DetectorElement& element ) const override;

  // Find the list of elements corresponding to a path (which can ba rehulare expression)
  StatusCode findElements( const std::string&                    path,
                           std::vector<const AlignmentElement*>& alignelements ) const override;

  // initialize the alignment frames now. if time=0, it will use the
  // current time. it will issue a warning, return failure and not do
  // anything if m_initTime != 0
  StatusCode initAlignmentFrame( Gaudi::Time now ) override;

  // return the time with which the geometry was initialized
  Gaudi::Time initTime() const override { return m_initTime; }
  // initialize an Al::Equations object
  void initEquations( Al::Equations& ) const override;

private:
  enum e_DoFs { Tx, Ty, Tz, Rx, Ry, Rz };

  void getElements( const IDetectorElement* parent, const RegExs& expressions, size_t depth,
                    std::vector<const DetectorElement*>& detelements ) const;

  struct ElementSorter {
    // this is very tricky: how do you sort elements in a tree if you
    // don't even know the tree( because of grouping). the following
    // should cover all cases (I hope).

    bool operator()( const AlignmentElement* lhs, const AlignmentElement* rhs ) const {
      return ( lhs->basename() < rhs->basename() ||
               ( lhs->basename() == rhs->basename() &&
                 ( lhs->detelements().front()->name() < rhs->detelements().front()->name() ||
                   ( lhs->detelements().front()->name() == rhs->detelements().front()->name() &&
                     lhs->elementsInTree().size() > rhs->elementsInTree().size() ) ) ) );

      // return lhs->basename() < rhs->basename() ||
      //	(lhs->basename() == rhs->basename() &&
      //	 lhs->detelements().front()->name() < rhs->detelements().front()->name() ) ;

      // all varant below do not work or do not give result I wanted.
      //       return lhs->detelements().front()->name() <
      // 	rhs->detelements().front()->name() ||
      // 	(lhs->detelements().front()->name() == rhs->detelements().front()->name() &&
      // lhs->elementsInTree().size() > rhs->elementsInTree().size() ) ;

      //       bool rhsIsLhsOffSpring = lhs->isOffspring( *rhs ) ;
      //       bool lhsIsRhsOffSpring = rhs->isOffspring( *lhs ) ;

      //       return (!rhsIsLhsOffSpring && !lhsIsRhsOffSpring) ?
      // 	lhs->detelements().front()->name() <
      // 	rhs->detelements().front()->name() :
      // 	rhsIsLhsOffSpring ;

      //       return lhs->elementsInTree().size() > rhs->elementsInTree().size() ||
      // 	(lhs->elementsInTree().size() == rhs->elementsInTree().size() &&
      // 	 lhs->detelements().front()->name() <
      // 	 rhs->detelements().front()->name() ) ;
      // anything based on the 'isOffspring' routine cannot work.
      // this does not work
      // return lhs->isOffspring( *rhs )  ||
      //( ! rhs->isOffspring( *lhs ) &&
      // lhs->detelements().front()->name() <
      // rhs->detelements().front()->name() ) ;
    }
  };

private:
  Gaudi::Property<bool> m_useLocalFrame{this, "UseLocalFrame", true};             ///< Use local frame as alignmentframe
  Gaudi::Property<std::vector<std::string>> m_elemsToBeAligned{this, "Elements"}; ///< Elemenst : Path to elements
  mutable IGetElementsToBeAligned::Elements m_elements;                           ///< Flat vector of alignment elements
  using ElementMap = std::map<const DetectorElement*, const AlignmentElement*>;
  ElementMap  m_elementMap; ///< Map of detector elements to alignment element
  Gaudi::Time m_initTime{0};
};
